class Sentence:
  """
  Utility parser to traverse and extract text fragments.
  """
  EOS = -1

  def __init__(self, value):
    """
    Creates a parser over a text.

    Args:
      value: Text to process.
    """
    self.value = value
    self.idx = 0

  def peek(self, value=0):
    """
    Returns the character at the current position + offset.

    Args:
      value: Relative offset from the current cursor.

    Returns:
      Character at the position or EOS if out of range.
    """
    if self.idx + value < len(self.value):
      return self.value[self.idx + value]
    return Sentence.EOS

  def next(self, value=1):
    """
    Advances the cursor N positions.

    Args:
      value: Number of positions to advance.
    """
    self.idx += value

  def peek_next(self):
    """
    Returns the current character and advances the cursor by 1.

    Returns:
      Current character or EOS if out of range.
    """
    ch = self.peek()
    self.next()
    return ch

  def is_eos(self):
    """
    Indicates whether the cursor is at the end of the text.

    Returns:
      True if no characters remain.
    """
    return self.idx >= len(self.value)


  def parse_until_word(self, word):
    """
    Extracts text until a word/token is found.

    Args:
      word: Expected ending token.

    Returns:
      Substring before the token or None if not found.
    """
    start = self.idx
    while not self.is_eos():
      if self.match(word):
        return self.value[start:self.idx - len(word)]
      self.next()
    return None


  def match(self, word):
    """
    Checks whether the text from the cursor matches a token.

    Args:
      word: Token to compare.

    Returns:
      True if it matches; advances the cursor in that case.
    """
    for f in range(len(word)):
      if self.peek(f) != word[f]:
        return False
    self.next(len(word))
    return True

  def substring(self, idx, idx2) -> str:
    """
    Returns a substring by absolute indices.

    Args:
      idx: Start index.
      idx2: End index (exclusive).

    Returns:
      Substring.
    """
    return self.value[idx:idx2]



# -*- coding: utf-8 -*-
# Catslap
# Author: Luis A. González
# MIT License (view LICENSE file)
# Copyright (c) 2026

# OOXML Python API by Luis Alberto González (SIA)
import sys
import json
import shutil

from catslap.base.document import Document
from catslap.base.relationships import Relationships
from catslap.base.types import ContentTypes
from catslap.pptx import elements as doc_elements
from catslap.xlsx.document import ExcelDocument, parse_data_ref, get_cell_format_position, get_cell_num, get_row_num
from catslap.utils import encoding as enc_util
from catslap.utils import file as file_util
from catslap.utils import html
from catslap.utils import text as text_util
from catslap.utils import types
from catslap.utils.xml import XmlParser, XmlTag, XmlText, CONFIG_PARAM_INCLUDE_DECL, CONFIG_PARAM_HTML

# -- puntos de powerpoint por cm
PT_PER_CM = 360000

PPT_SLIDES = "ppt/slides/"
PPT_DIAGRAMS = "ppt/diagrams/"
PPT_EMBEDINGS = "ppt/embeddings/"
PPT_CHARTS = "ppt/charts/"
PPT_CHARTS_RELS = PPT_CHARTS + "_rels/"
DOT_RELS = '.rels'

WORD_DOCUMENT_RELS = "word/_rels/document.xml.rels"
PPT_DOCUMENT_TYPES = "[Content_Types].xml"

TAG_SLD = 'p:sld'
TAG_DATAMODEL = 'dgm:dataModel'
TAG_TBL = 'a:tbl'
TAG_TR = 'a:tr'
TAG_P = 'a:p'
TAG_R = 'a:r'
TAG_RPR = 'a:rPr'
TAG_T = 'a:t'
TAG_SOLID_FILL = 'a:solidFill'
TAG_NO_FILL = 'a:noFill'
TAG_SRGB_CLR = 'a:srgbClr'


IGNORABLE_TAGS = ['c:lang']
IGNORABLE_EMPTY_TAGS = ['a:effectLst']
IGNORABLE_EMPTY_STYLE_TAGS = ['a:ea', 'a:cs', TAG_NO_FILL, 'a:ln', 'a:uLnTx', 'a:uFillTx', 'a:latin']
IGNORABLE_ATTRS = ['dirty', 'err']


class PowerPointDocument(Document):
  """
  Processes and generates PowerPoint presentations (.pptx).

  Attributes:
    max_id: Maximum ID used in the document.
    types: Package ContentTypes.
    relationships: Relationships for the current slide.
  """
  def __init__(self, file: str):
    super().__init__(file)
    self.max_id = 0
    self.types = None
    self.relationships = None

  def process_template(self, tempdir: str):
    """
    Processes the PPTX template in the temporary directory.

    Args:
      tempdir: Temporary directory with ZIP contents.
    """
    ppt_types = tempdir + "/" + PPT_DOCUMENT_TYPES
    self.types = ContentTypes(ppt_types)
    self.__process_ppt_slides(tempdir, TAG_SLD)
    if not self.test_mode:
      self.types.write_file()

  def __process_ppt_slides(self, tempdir: str, tag_name: str):
    xml = XmlParser()
    # -- parsea los slides de la PPT
    idx = 1
    found = True
    while found:
      slide_file = tempdir + '/' + PPT_SLIDES + "slide" + str(idx) + ".xml"
      found = file_util.exist(slide_file)
      if found:
        file0_rel = tempdir + '/' + PPT_SLIDES + "_rels/slide" + str(idx) + ".xml.rels"
        self.relationships = Relationships(tempdir, file0_rel)
        document = xml.parse_file(slide_file, tag_name)
        blocks = document.elements
        self.collapse_paragraphs(blocks)
        self.max_id = max(self.max_id, xml.max_id)
        self.search_graphic_frames(tempdir, blocks)
        self.process_descr_attrs(blocks)
        self.process_paragraphs(blocks)
        self.process_html_content(blocks)
        xml_content = xml.get_pretty_xml(document, {CONFIG_PARAM_INCLUDE_DECL: True})
        file_util.write_bytes(slide_file, bytes(xml_content, enc_util.UTF_8))
        if not self.test_mode:
          self.relationships.write_file()
        idx += 1
    idx = 1
    found = True
    while found:
      diagram_file = tempdir + '/' + PPT_DIAGRAMS + "data" + str(idx) + ".xml"
      found = file_util.exist(diagram_file)
      if found:
        self.relationships = None
        document = xml.parse_file(diagram_file, TAG_DATAMODEL)
        blocks = document.elements
        self.collapse_paragraphs(blocks)
        self.max_id = max(self.max_id, xml.max_id)
        self.process_paragraphs(blocks)
        self.process_html_content(blocks)
        xml_content = xml.get_pretty_xml(document, {CONFIG_PARAM_INCLUDE_DECL: True})
        file_util.write_bytes(diagram_file, bytes(xml_content, enc_util.UTF_8))
        idx += 1

    files = file_util.list_files(tempdir + '/' + PPT_CHARTS_RELS)
    if files:
      for file in files:
        chart_file = tempdir + '/' + PPT_CHARTS_RELS + file
        if file.endswith('.xml.rels'):
          chart_name = file[0:len(file) - 5]
          relationships = Relationships(tempdir, chart_file)
          relations = relationships.get_relationships(None, ".xlsx")
          if len(relations) == 0:
            continue
          relation = relations[0]
          target = relation.target
          self.__process_chart_excel_file(chart_name, tempdir, target)

  def __process_chart_excel_file(self, chart_name, tempdir, target):
    excel = ExcelDocument(target, False)
    try:
      if self.test_mode:
        excel.test_with_json({})
      else:
        excel.create_doc_with_resolvers(target, self.value_resolver, self.repeating_resolver)
        chart_file = tempdir + '/' + PPT_CHARTS + chart_name
        self.clear_chart_values(chart_file, excel)
    finally:
      excel.close()
    self.access_ok_param_list = types.merge_list_unique(self.access_ok_param_list, excel.access_ok_param_list)
    self.access_err_param_list = types.merge_list_unique(self.access_err_param_list, excel.access_err_param_list)

  def clear_chart_values(self, chart_file, excel):
    """
    Updates and clears chart series caches.

    Args:
      chart_file: Chart file.
      excel: Auxiliary Excel document.
    """
    if self.test_mode:
      return
    parser = XmlParser()
    tag = parser.parse_file(chart_file, 'c:chartSpace')
    self.collapse_paragraphs(tag.elements)
    tag = tag.get_tag_path(['c:chart', 'c:plotArea', '*Chart'])
    sers = tag.get_tags('c:ser')
    for ser in sers:
      cat = ser.get_tag('c:cat')
      ref_tag = cat.get_tag('c:strRef', False)
      tag_name = 'c:strCache'
      if ref_tag is None:
        ref_tag = cat.get_tag('c:numRef', False)
        tag_name = 'c:numCache'
      if ref_tag is not None:
        f_tag = ref_tag.get_tag('c:f')
        sheet_ref = f_tag.get_text()
        str_cache = ref_tag.get_tag(tag_name)
        str_cache.clear_tags()
        # -- extrae los datos desde la referencia de excel (Ej: Hoja1!$A$2:$A$3)
        sheet_name, sdata, edata = parse_data_ref(sheet_ref)
        scell_num = get_cell_num(sdata)
        srow_num = get_row_num(sdata)
        ecell_num = get_cell_num(edata)
        sdata = get_cell_format_position(scell_num, srow_num, True)
        sheet_ref = sheet_name + "!" + sdata
        values = excel.extract_data(sheet_ref)
        edata = get_cell_format_position(ecell_num, srow_num + len(values) - 1, True)
        sdata = get_cell_format_position(scell_num, srow_num, True)
        sheet_ref = sheet_name + "!" + sdata + ":" + edata
        f_tag.set_text(sheet_ref)
        str_cache.add_tag('c:ptCount', {'val': len(values)})
        for idx in range(0, len(values)):
          value_idx = values[idx]
          if value_idx and len(value_idx) > 0:
            str_cache.add_tag('c:pt', {'idx': idx}).add_tag_text('c:v', value_idx[0])
      vals = ser.get_tags('c:val')
      for val in vals:
        ref_tag = val.get_tag('c:numRef', False)
        tag_name = 'c:numCache'
        if ref_tag is None:
          ref_tag = val.get_tag('c:strRef', False)
          tag_name = 'c:strCache'
        if ref_tag is not None:
          f_tag = ref_tag.get_tag('c:f')
          sheet_ref = f_tag.get_text()
          sheet_name, sdata, edata = parse_data_ref(sheet_ref)
          scell_num = get_cell_num(sdata)
          srow_num = get_row_num(sdata)
          ecell_num = get_cell_num(edata)
          sdata = get_cell_format_position(scell_num, srow_num, True)
          sheet_ref = sheet_name + "!" + sdata
          values = excel.extract_data(sheet_ref)
          edata = get_cell_format_position(ecell_num, srow_num + len(values) - 1, True)
          sdata = get_cell_format_position(scell_num, srow_num, True)
          sheet_ref = sheet_name + "!" + sdata + ":" + edata
          f_tag.set_text(sheet_ref)

          num_cache = ref_tag.get_tag(tag_name)
          num_cache.clear_tags()
          # -- extrae los datos desde la referencia de excel
          values = excel.extract_data(sheet_ref)
          num_cache.add_tag_text('c:formatCode', 'General')
          num_cache.add_tag('c:ptCount', {'val': len(values)})
          for idx in range(0, len(values)):
            value_idx = values[idx]
            if value_idx and len(value_idx) > 0:
              num_cache.add_tag('c:pt', {'idx': idx}).add_tag_text('c:v', value_idx[0])
    parser.write_file()

  def collapse_paragraphs(self, elements: list, rep: int = 0):
    """
    Collapses runs and removes ignorable tags in an XML tree.

    Args:
      elements: List of XML elements.
      rep: Recursion level (internal use).
    """
    idx = 0
    last_rpr = None
    last_t = None
    while idx < len(elements):
      element = elements[idx]
      if not isinstance(element, XmlTag):
        idx += 1
        continue
      self.collapse_paragraphs(element.elements, rep + 1)
      tag = element
      for attr_name in IGNORABLE_ATTRS:
        tag.remove_attr(attr_name)
      tag_name = tag.name
      tags = tag.elements
      # -- ignora el tag
      if tag_name in IGNORABLE_TAGS or (len(tags) == 0 and tag_name in IGNORABLE_EMPTY_TAGS):
        del elements[idx]
        continue
        # -- funde todos los a:r que sean iguales en un solo a:r
      if tag_name == TAG_R:
        rpr = tag.get_tag(TAG_RPR, False)
        t = tag.get_tag(TAG_T, False)
        if len(tags) == 0 or t is None:
          idx += 1
          continue
        tag_name = t.name
        if tag_name == TAG_T:
          if last_t is not None and PowerPointDocument.__is_the_same_rpr(last_rpr, rpr):
            ct = t.elements[0].content if len(t.elements) > 0 and isinstance(t.elements[0], XmlText) else None
            if ct is not None:
              if len(last_t.elements) == 0:
                last_t.add_text(ct)
              else:
                last_t.elements[0].append(ct)
            last_t.attrs['xml:space'] = 'preserve'
            del elements[idx]
            continue
          last_rpr = rpr
          last_t = t
        else:
          last_rpr = None
          last_t = None
      idx += 1

  @staticmethod
  def __is_the_same_rpr(rpr1: XmlTag or None, rpr2: XmlTag or None) -> bool:
    if rpr1 is None and rpr2 is None:
      return True
    xml = XmlParser()
    rpr1 = rpr1.clone(True)
    rpr2 = rpr2.clone(True)
    PowerPointDocument.__remove_ignorable_style_tags(rpr1)
    PowerPointDocument.__remove_ignorable_style_tags(rpr2)
    dump1 = xml.get_outer_xml(rpr1)
    dump2 = xml.get_outer_xml(rpr2)
    return dump1 == dump2

  @staticmethod
  def __remove_ignorable_style_tags(tag: XmlTag):
    elements = tag.elements
    idx = 0
    while idx < len(elements):
      element = elements[idx]
      if not isinstance(element, XmlTag):
        idx += 1
        continue
      tag_name = element.name
      if tag_name in IGNORABLE_EMPTY_STYLE_TAGS:
        del elements[idx]
        continue
      PowerPointDocument.__remove_ignorable_style_tags(element)
      idx += 1

  def __resolve_descr(self, descr):
    descr = XmlParser.resolve_entities(descr)
    descr = text_util.trim(descr)
    if not descr.startswith("{{") or not descr.endswith("}}"):
      return ''
    param = descr[2:len(descr) - 2]
    value = self.resolve_value(None, param)
    if text_util.is_empty(value):
      return ''
    if isinstance(value, dict):
      value = json.dumps(value)
    elif isinstance(value, str) and value.startswith('{'):
      value = eval(value)
    return value

  def search_graphic_frames(self, tempdir, elements0):
    """
    Finds and processes graphic frames in slides.

    Args:
      tempdir: Temporary directory.
      elements0: List of XML elements.
    """
    idx0 = 0
    while idx0 < len(elements0):
      element0 = elements0[idx0]
      if isinstance(element0, XmlTag):
        if element0.name == 'p:graphicFrame':
          self.__process_graphic_frame(tempdir, element0)
        else:
          self.search_graphic_frames(tempdir, element0.elements)
      idx0 += 1

  def __process_graphic_frame(self, tempdir, element0):
    cnvpr = element0.get_tag_path(['p:nvGraphicFramePr', 'p:cNvPr'])
    chart = element0.get_tag_path(['a:graphic', 'a:graphicData', 'c:chart'], False)
    if cnvpr is None or chart is None:
      return
    descr = cnvpr.get_attr('descr')
    if descr is None:
      return
      # -- obtiene los datos de la inyección
    value = self.__resolve_descr(descr)
    if not value.startswith('{') or self.test_mode:
      return
    cnvpr.set_attr('descr', '')

    rid = chart.get_attr('r:id')
    if rid is None:
      return
    rel = self.relationships.get_relationship_by_id(rid)
    if rel is None:
      return
    target = rel.target
    tfile = file_util.get_filename(target)
    chart_rel_file = tempdir + '/' + PPT_CHARTS_RELS + tfile + DOT_RELS
    relationships = Relationships(tempdir, chart_rel_file)
    relations = relationships.get_relationships(None, ".xlsx")
    if len(relations) == 0:
      return
    relation = relations[0]
    # -- obtiene nombre del archivo excel
    xml_file = relation.target
    # -- carga el json
    diagram = json.loads(value)
    # Crear datos en las celdas A1:C5
    datos = [
            diagram.get('legends')
        ]
    cats = diagram.get('categories')
    series = diagram.get('series')
    for idx, cat in enumerate(cats):
      row = [cat]
      for serie in series:
        data = serie.get('data')
        value = data[idx] if idx < len(data) else 0
        row.append(value)
      datos.append(row)

    path = file_util.get_pathname(sys.modules['ooxml2.xlsx'].__file__)
    path = path + 'empty.xlsx'
    shutil.copy2(path, xml_file)
    # -- abre archivo excel para escritura
    excel = ExcelDocument(xml_file, False)
    try:
      excel.write_cells(datos)
      dbytes = excel.get_document_bytes()
      file_util.write_bytes(xml_file, dbytes)
      chart_file = tempdir + '/' + PPT_CHARTS + tfile
      self.clear_chart_values(chart_file, excel)
    finally:
      excel.close()
      # -- modifica charts
    chart_file = tempdir + '/' + PPT_CHARTS + tfile
    parser = XmlParser()
    root_tag = parser.parse_file(chart_file)
    title_tag = root_tag.get_tag_path(['c:chart', 'c:title', 'c:tx', 'c:rich', 'a:p', 'a:r', 'a:t'], False)
    if title_tag:
      title_tag.set_text(diagram.get('title'))
    chart_tag = root_tag.get_tag_path(['c:chart', 'c:plotArea', '*Chart'], False)
    if chart_tag is not None:
      ser_tags = chart_tag.get_tags('c:ser')
      for idx, ser_tag in enumerate(ser_tags):
        if idx >= len(series):
          chart_tag.remove(ser_tag)
          continue
        v_tag = ser_tag.get_tag_path(['c:tx', 'c:strRef', 'c:strCache', 'c:pt', 'c:v'], False)
        if v_tag:
          name = series[idx].get('name')
          if name:
            v_tag.set_text(name)
    parser.write_file()

  def process_descr_attrs(self, elements0):
    """
    Processes shape 'descr' attributes and replaces content.

    Args:
      elements0: List of XML elements.
    """
    idx0 = 0
    while idx0 < len(elements0):
      element0 = elements0[idx0]
      if not isinstance(element0, XmlTag):
        idx0 += 1
        continue
      elements = element0.elements
      self.process_descr_attrs(elements)
      for element in elements:
        if not isinstance(element, XmlTag):
          continue
        if element.name != 'p:cNvPr':
          continue
        descr = element.get_attr('descr')
        if descr is None:
          continue
        value = self.__resolve_descr(descr)
        element.set_attr('descr', '')
        if isinstance(value, dict):
          shape = value
          if shape is None:
            continue
          shape = dict(shape)
          x = shape.get('x')
          y = shape.get('y')
          wd = shape.get('wd')
          hg = shape.get('hg')
          bg = shape.get('bg')
          fg = shape.get('fg')
          txt = shape.get('text')
        else:
          x = None
          y = None
          wd = None
          hg = None
          bg = None
          fg = None
          txt = str(value)
          # -- busca posición de objeto
        idx1 = 0
        sppr = None
        txbody = None
        while idx1 < len(elements0):
          element1 = elements0[idx1]
          if element1.name == 'p:spPr':
            sppr = element1
          if element1.name == 'p:txBody':
            txbody = element1
          idx1 += 1
          # -- posición, dimensión y color de fondo
        if sppr is not None:
        # -- busca posición de la forma
          xfrm = sppr.get_tag('a:xfrm', False)
          if xfrm is not None:
            off = xfrm.get_tag('a:off')
            ext = xfrm.get_tag('a:ext')
            if x is not None and text_util.is_decimal(x):
              x_value = int(round(PT_PER_CM * float(x)))
              off.set_attr('x', x_value)
            if y is not None and text_util.is_decimal(y):
              y_value = int(round(PT_PER_CM * float(y)))
              off.set_attr('y', y_value)
            if wd is not None and text_util.is_decimal(wd):
              wd_value = int(round(PT_PER_CM * float(wd)))
              ext.set_attr('cx', wd_value)
            if hg is not None and text_util.is_decimal(hg):
              hg_value = int(round(PT_PER_CM * float(hg)))
              ext.set_attr('cy', hg_value)
              # -- busca color de la forma
          solid_fill = sppr.get_tag(TAG_SOLID_FILL, False)
          if solid_fill is None:
            solid_fill = sppr.get_tag(TAG_NO_FILL, False)
            if solid_fill is not None:
              solid_fill.name = TAG_SOLID_FILL
          if solid_fill is not None and bg is not None and len(bg) == 6 and text_util.is_hex(bg):
            solid_fill.clear_tags()
            solid_fill.add_tag(TAG_SRGB_CLR, {'val': bg})
            # -- color de texto
        if txbody is not None and fg is not None and len(fg) == 6 and text_util.is_hex(fg):
          def __change_solid_fill_tag_color(tag, rgb) -> bool:
            if tag.name == TAG_NO_FILL:
              tag.name = TAG_SOLID_FILL
            if tag.name == TAG_SOLID_FILL:
              tag.clear_tags()
              tag.add_tag(TAG_SRGB_CLR, {'val': rgb})
              return True
            for elem in tag.elements:
              if __change_solid_fill_tag_color(elem, rgb):
                return True
            return False
          __change_solid_fill_tag_color(txbody, fg)
          # -- sustitucion de texto
        if txbody is not None and txt is not None:
          tag_p = txbody.get_tag('a:p', False)
          tag_r = tag_p.get_tag('a:r', False) if tag_p is not None else None
          if tag_r is None:
            tag_endrpr = tag_p.get_tag('a:endParaRPr', False)
            if tag_endrpr is not None:
              tag_p.remove_tag('a:endParaRPr')
              tag_r = tag_p.add_tag('a:r')
              tag_endrpr.name = 'a:rPr'
              tag_r.add_element(tag_endrpr)
            else:
              tag_r = tag_p.add_tag('a:r')
          tag_t = tag_r.get_tag('a:t', False)
          if tag_t is None:
            tag_t = tag_r.add_tag('a:t')
          txt = text_util.trim(txt)
          tag_t.set_text(XmlParser.escape_entities(txt))
      idx0 += 1

  def process_paragraphs(self, elements: list):
    """
    Processes paragraphs resolving placeholders and repetitions.

    Args:
      elements: List of XML elements.
    """
    idx = 0
    while idx < len(elements):
      element = elements[idx]
      if not isinstance(element, XmlTag):
        idx += 1
        continue
      tag_name = element.name
      # -- bloques de primer nivel (párrafos, tablas, etc.)
      # -- Caso 4: hay repeticiones dentro de una tabla (repeticiones de fila)
      if tag_name == TAG_TBL:
        self.__process_table(element)
        if len(element.elements) == 0:
          del elements[idx]
          continue
        idx += 1
        continue
      sometext0, somedollar0, _ = self.__resolve_text_value(None, elements[idx])
      if not sometext0 and somedollar0:
        del elements[idx]
        continue
      idx += 1

  def __process_table(self, tbl_tag: XmlTag):
    tr_tags = tbl_tag.elements
    pos = 0
    while pos < len(tr_tags):
      tr_tag = tr_tags[pos]
      # -- asegura que se procesan las filas de la tabla
      if not isinstance(tr_tag, XmlTag) or tr_tag.name != TAG_TR:
        pos += 1
        continue
        # -- comprueba si hay repeticiones de fila dentro de las celdas
      repeating = self.__resolve_text_repeating(tr_tag)
      if repeating > 1:
        self. __repeat_block_from(tr_tags, pos, repeating)
        continue
        # -- borra la fila si no hay texto o si hay pero no es de dollar habiendo dollar
      sometext, somedollar, sometextdollar = self.__resolve_text_value(None, tr_tags[pos])
      if not sometext or (somedollar and not sometextdollar):
        del tr_tags[pos]
        continue
      pos += 1

  def __repeat_block_from(self, tags: list, idx: int, repeating: int):
    """
    Repeats a tag from a tag list at a given position.
    """
    newblocks = []
    elem = tags[idx]
    for row in range(0, repeating):
      tr0 = elem.clone()
      sometext, somedollar, sometextdollar = self.__resolve_text_value(row, tr0)
      # si es un tr con $ en alguna celda pero sin texto de $, ignora toda la fila
      if not sometextdollar and somedollar:
        continue
      self.max_id = PowerPointDocument.__reassign_ids(tr0, self.max_id)
      newblocks.append(tr0)
      # -- elimina el bloque que sirvió de patrón
    del tags[idx]
    # -- inserta todos los nuevos bloques resueltos
    pos = idx
    for newblock in newblocks:
      tags.insert(pos, newblock)
      pos += 1

  @staticmethod
  def __reassign_ids(block: XmlTag, maxid: int) -> int:
    if not isinstance(block, XmlTag):
      return maxid
    elements = block.elements
    for item in elements:
      maxid = PowerPointDocument.__reassign_ids(item, maxid)
    attrs = block.attrs
    if not attrs or len(attrs) == 0:
      return maxid
    attrname = 'o:spid'
    idvalue = attrs.get(attrname)
    if idvalue:
      del attrs[attrname]
    attrname = 'id'
    idvalue = attrs.get(attrname)
    if not idvalue:
      return maxid
    try:
      maxid += 1
      int(idvalue)
      attrs[attrname] = maxid
    except ValueError:
      attrs[attrname] = '_id_n' + str(maxid)
    return maxid

  def __resolve_text_value(self, row: int or None, block: XmlTag) -> (bool, bool, bool):
    """
    Processes a tag and resolves its text when present.
    """
    if not isinstance(block, XmlTag):
      return False, False, False
    tag = block.name
    elements = block.elements
    if tag != TAG_T:
      sometext = False
      somedollar = False
      sometextdollar = False
      idx = 0
      while idx < len(elements):
        item = elements[idx]
        if not isinstance(item, XmlTag):
          idx += 1
          continue
        sometext0, somedollar0, sometextdollar0 = self.__resolve_text_value(row, item)
        tag0 = item.name
        # -- si un TR no tienen entre sus TC algo de texto resuelto con $, se borra
        if tag0 == TAG_TR and not sometextdollar0 and somedollar0:
          del elements[idx]
          continue
          # -- si un TBL no tienen entre sus TC algo de texto resuelto con $, se borra
        if tag0 == TAG_TBL and not sometextdollar0 and somedollar0:
          del elements[idx]
          continue
        sometext = sometext0 or sometext
        somedollar = somedollar0 or somedollar
        sometextdollar = sometextdollar0 or sometextdollar
        idx += 1
      return sometext, somedollar, sometextdollar
    text_node = elements[0] if len(elements) > 0 else None
    if not isinstance(text_node, XmlText):
      return False, False, False
    value = text_node.content
    if not isinstance(value, str):
      return False, False, False
    idx1 = value.find('{{')
    if idx1 < 0:
      return text_util.trim(value) != '', False, False
    htext = ''
    text2 = ''
    idx0 = 0
    while idx1 >= 0:
      idx2 = value.find('}}', idx1)
      if idx2 <= idx1:
        break
      htext += value[idx0:idx1]
      text2 += value[idx0:idx1]
      param = value[idx1+2:idx2]
      resolved = self.resolve_value(row, param)
      if resolved is not None:
        htext = htext + str(resolved)
      idx0 = idx2 + 2
      idx1 = value.find('{{', idx0)
    htext = text_util.trim(htext + value[idx0:])
    text2 = text_util.trim(text2 + value[idx0:])
    sometext = htext != text2
    if sometext:
      text1 = PowerPointDocument.normalize_html_text(htext)
      text_node.content = text1
    else:
      text_node.content = ''
    return sometext, True, sometext

  def __resolve_text_repeating(self, block: XmlTag) -> int:
    if not isinstance(block, XmlTag):
      return 0
    tag = block.name
    elements = block.elements
    if tag != TAG_T:
      repeating = 0
      for item in elements:
        repeating = max(self.__resolve_text_repeating(item), repeating)
      return repeating
    text_node = elements[0]
    if not isinstance(text_node, XmlText):
      return 0
    value = text_node.content
    idx2 = 0
    repeating = 0
    while idx2 >= 0:
      idx1 = value.find('{{', idx2)
      idx2 = value.find('}}', idx1) if idx1 >= 0 else -1
      if idx2 > idx1:
        param = value[idx1 + 2:idx2]
        rep0 = self.resolve_repeating(param)
        if rep0 is None:
          rep0 = 0
        repeating = max(repeating, rep0)
        idx2 += 2
    return repeating

  @staticmethod
  def __find_block(tagname: str, block: XmlTag):
    if not isinstance(block, XmlTag):
      return None
    tag = block.name
    if tag == tagname:
      return block
    content = block.elements
    for item in content:
      found = PowerPointDocument.__find_block(tagname, item)
      if found:
        return found
    return None

  def process_html_content(self, blocks: list):
    """
    Expands embedded HTML content in text.

    Args:
      blocks: List of XML elements.
    """
    idx = 0
    while idx < len(blocks):
      block = blocks[idx]
      if not isinstance(block, XmlTag):
        idx += 1
        continue
      tag = block.name
      if tag == TAG_P:
        rblock = PowerPointDocument.__find_block(TAG_R, block)
        if rblock:
          r_elements = rblock.elements
          if len(r_elements) == 0:
            idx += 1
            continue
          t_tag = None
          for rtag in r_elements:
            if not isinstance(rtag, XmlTag):
              continue
            tag = rtag.name
            if tag == TAG_T:
              t_tag = rtag
              break
          if not t_tag or len(t_tag.elements) == 0 or not isinstance(t_tag.elements[0], XmlText):
            self.process_html_content(r_elements)
            idx += 1
            continue
          tcontent = t_tag.elements[0].content
          if tcontent.find('<') < 0:
            idx += 1
            continue
          html_blocks = self.__expand_html_content(tcontent)
          if html_blocks:
            del blocks[idx]
            for html_block in html_blocks:
              blocks.insert(idx, html_block)
              idx += 1
          else:
            idx += 1
          continue
      content = block.elements
      if isinstance(content, list):
        self.process_html_content(content)
      idx += 1

  def __expand_html_content(self, text_html: str) -> list or None:
    out = []
    xml = XmlParser({CONFIG_PARAM_HTML: True})
    blocks = xml.parse_text(text_html)
    for block in blocks:
      if not isinstance(block, XmlTag):
        continue
      tag = block.name
      attrs = block.attrs
      paraprops = {
                'align': None,
                'size': 800,
                'codeblock': None
            }
      classname = attrs.get('class')
      if classname:
      # -- asegura que la clase no tiene el caret
        classname = text_util.trim(classname.replace('caret', ''))
        # -- mapea las classes HTML a estilos WORD
        classes = classname.split(' ')
        for classname in classes:
          if classname == 'left':
            paraprops['align'] = 'l'
          elif classname == 'right':
            paraprops['align'] = 'r'
          elif classname == 'center':
            paraprops['align'] = 'ctr'
          elif classname == 'justify':
            paraprops['align'] = 'just'
          elif classname.startswith('list-'):
            paraprops['list'] = classname[5:]
          elif classname == 'codeblock' or classname == 'token':
            paraprops['codeblock'] = True
          elif classname == 'security-level':
            paraprops['align'] = 'l'
            paraprops['bold'] = True
            paraprops['size'] = 1000
          elif classname == 'link-title':
            paraprops['align'] = 'l'
            paraprops['bold'] = True
            paraprops['size'] = 800
          elif classname == 'link-url':
            paraprops['align'] = 'l'
            paraprops['code'] = True
            paraprops['size'] = 800

      elements = block.elements
      astyle = attrs.get('style')
      if astyle:
        css = html.parse_css(astyle)
        PowerPointDocument.parse_css_properties(css, paraprops)
      if tag == 'pre':
        paraprops['codeblock'] = True
      if tag == 'ul' or tag == 'ol':
        for li in elements:
          if not isinstance(li, XmlTag):
            continue
          self.__create_paragraph(out, paraprops, li.elements)
        continue
      headers = ['h1', 'h2', 'h3', 'h4', 'h5', 'h6']
      try:
        idx = headers.index(tag)
      except ValueError:
        idx = -1
      if idx >= 0:
        paraprops['size'] = 800 + ((5 - idx) * 200)
        self.__create_paragraph(out, paraprops, elements)
        continue
      self.__create_paragraph(out, paraprops, elements)
    return out

  def __create_paragraph(self, out: list, paraprops: dict, content: str or list):
    if not isinstance(content, list):
      out.append(doc_elements.create_paragraph(paraprops, content, self.relationships, self.types))
      return
    properties = paraprops.copy()
    if properties.get('codeblock'):
      properties['code'] = True
    runs = []
    self.__process_tag_content(content, runs, properties)
    out.append(doc_elements.create_paragraph(paraprops, runs, self.relationships, self.types))

  def __process_tag_content(self, blocks: list, runs: list, properties: dict):
    for item in blocks:
      if isinstance(item, XmlText):
        runs.append(doc_elements.create_run(item.content, properties, self.relationships, self.types))
        continue
      tag = item.name
      props = properties.copy()
      attrs = item.attrs
      if tag == 'font':
        color = attrs.get('color')
        if color:
          props['color'] = color
      elif tag == 'b':
        props['bold'] = True
      elif tag == 'i':
        props['italic'] = True
      elif tag == 'strike' or tag == 'stroke':
        props['strike'] = True
      elif tag == 'u':
        props['underline'] = True
      elif tag == 'code':
        props['code'] = True
      elif tag == 'a':
        url = attrs.get('href')
        if url:
          props['link'] = url
          props['color'] = '#0000ff'
      elif tag == 'img':
        pass  # -- No se soportan imágenes en línea de párrafo
      astyle = attrs.get('style')
      if astyle:
        css = html.parse_css(astyle)
        PowerPointDocument.parse_css_properties(css, props)

      content = item.elements
      self.__process_tag_content(content, runs, props)

  @staticmethod
  def parse_css_properties(css: dict or None, props):
    """
    Applies CSS properties to a props dictionary.

    Args:
      css: CSS dictionary.
      props: Destination properties dictionary.
    """
    if not css:
      return
    text_align = css.get('text-align')
    if text_align:
      align = None
      if text_align == 'left':
        align = 'l'
      elif text_align == 'right':
        align = 'r'
      elif text_align == 'center':
        align = 'ctr'
      elif text_align == 'justify':
        align = 'just'
      if align:
        props['align'] = align
    font_weight = css.get('font-weight')
    if font_weight:
      props['bold'] = True if font_weight != 'normal' and font_weight != '200' else False
    font_style = css.get('font-style')
    if font_style:
      props['italic'] = True if font_style == 'italic' else False
    color = css.get('color')
    if color:
      color = html.get_rgb_color(color)
      props['color'] = color
    text_decoration = css.get('text-decoration')
    if text_decoration:
      props['underline'] = False
      props['strike'] = False
      if text_decoration == 'underline':
        props['underline'] = True
      elif text_decoration == 'line-through':
        props['strike'] = True
    font_size = css.get('font-size')
    if font_size and font_size.endswith("px"):
      try:
        font_size = int(text_util.trim(font_size[0:len(font_size) - 2])) * 100
        props['size'] = font_size
      except ValueError:
        pass
    pxhg = css.get('height')
    if pxhg and pxhg.endswith("px"):
      try:
        pxhg = int(text_util.trim(pxhg[0:len(pxhg) - 2]))
        props['height'] = pxhg
      except ValueError:
        pass
    pxwd = css.get('width')
    if pxwd and pxwd.endswith("px"):
      try:
        pxwd = int(text_util.trim(pxwd[0:len(pxwd) - 2]))
        props['width'] = pxwd
      except ValueError:
        pass



# -*- coding: utf-8 -*-
# Catslap
# Author: Luis A. González
# MIT License (view LICENSE file)
# Copyright (c) 2026

import os
import shutil
import tempfile
from pathlib import Path


def get_temp_dir():
  """
  Creates a temporary directory.

  Returns:
    Path to the created temporary directory.

  Raises:
    OSError: If the temporary directory cannot be created.
  """
  return tempfile.mkdtemp()


def get_pathname(path: str):
  """
  Gets the directory portion (with trailing slash) of a path.

  Args:
    path: File or directory path.

  Returns:
    Directory portion ending with '/' or an empty string if not present.
  """
  idx = path.rfind('/')
  if idx > 0:
    path = path[0:idx + 1]
    return path
  return ''


def get_filename(path: str):
  """
  Gets the filename without its directory.

  Args:
    path: Full path.

  Returns:
    Filename.
  """
  idx = path.rfind('/')
  if idx >= 0:
    return path[idx + 1:]
  return path


def get_extension(path: str):
  """
  Gets a file extension (including the dot).

  Args:
    path: File path.

  Returns:
    Extension including the dot or an empty string if missing.
  """
  idx = path.rfind('.')
  if idx >= 0:
    return path[idx:]
  return ''


def exist(file):
  """
  Indicates whether a path exists as a file or directory.

  Args:
    file: Path to check.

  Returns:
    True if it exists as a file or directory.
  """
  return is_file(file) or is_directory(file)

def is_directory(file):
  """
  Indicates whether a path is a directory.

  Args:
    file: Path to check.

  Returns:
    True if it is a directory.
  """
  return os.path.isdir(file)

def is_file(file):
  """
  Indicates whether a path is a file.

  Args:
    file: Path to check.

  Returns:
    True if it is a file.
  """
  return os.path.isfile(file)

def get_base_dir(file):
  """
  Gets the base directory of the current module.

  Args:
    file: Ignored parameter (kept for compatibility).

  Returns:
    Directory path of the current file.
  """
  return Path(__file__).resolve().parent.as_posix()

def complete_path(path: str, relpath: str) -> str:
  """
  Resolves a relative path against a base path.

  Args:
    path: Base path.
    relpath: Relative path (supports ./ and ../).

  Returns:
    Absolute/combined path.

  Raises:
    RecursionError: If the relative path triggers infinite recursion (edge case).
  """
  if not path.endswith('/'):
    path = path + '/'
  if relpath.startswith('/'):
    return path + complete_path('/', relpath[1:])[1:]
  if relpath.startswith('./'):
    return complete_path(path, relpath[2:])
  if relpath.startswith('../'):
    if path.endswith('/'):
      path = path[0:len(path)-1]
    path = get_pathname(path)
    return complete_path(path, relpath[3:])
  idx = relpath.find('/')
  if idx > 0:
    return complete_path(path + relpath[0:idx + 1], relpath[idx + 1:])
  return path + relpath


def get_temp_file(tmpdir: str | None, extension: str) -> str:
  """
  Creates a temporary file and returns its path.

  Args:
    tmpdir: Temporary directory (None to use system default).
    extension: Temporary file suffix.

  Returns:
    Path to the created temporary file.

  Raises:
    OSError: If the file cannot be created.
  """
  (handle, name) = tempfile.mkstemp(suffix=extension, prefix=None, dir=tmpdir, text=False)
  os.close(handle)
  return name


def read_bytes(file: str) -> bytes:
  """
  Reads the binary contents of a file.

  Args:
    file: File path.

  Returns:
    Read bytes.

  Raises:
    FileNotFoundError: If the file does not exist.
    OSError: If read errors occur.
  """
  handler = open(file, "rb")
  with handler:
    return handler.read()


def read_text(file: str, encoding: str) -> str:
  """
  Reads a text file with a specific encoding.

  Args:
    file: File path.
    encoding: Text encoding.

  Returns:
    Read text.

  Raises:
    FileNotFoundError: If the file does not exist.
    UnicodeDecodeError: If decoding fails.
    OSError: If read errors occur.
  """
  data = read_bytes(file)
  return data.decode(encoding)


def write_bytes(file: str, data: bytes):
  """
  Writes bytes to a file.

  Args:
    file: File path.
    data: Binary content to write.

  Raises:
    OSError: If write errors occur.
  """
  infile_handler = open(file, "wb")
  with infile_handler:
    infile_handler.write(data)


def remove_dir_tree(directory: str):
  """
  Removes a directory recursively.

  Args:
    directory: Directory path.

  Raises:
    FileNotFoundError: If the directory does not exist.
    OSError: If it cannot be removed.
  """
  shutil.rmtree(directory)


def remove_file(filename: str):
  """
  Removes a file if it exists.

  Args:
    filename: File path.

  Raises:
    OSError: If it cannot be removed.
  """
  if os.path.exists(filename):
    os.remove(filename)


def list_files(directory: str) -> list | None:
  """
  Lists files and directories in a directory.

  Args:
    directory: Directory path.

  Returns:
    List of names or None if it does not exist.

  Raises:
    OSError: If a directory read error occurs.
  """
  try:
    return os.listdir(directory)
  except FileNotFoundError:
    return None


def copy_to_temp_dir(filename: str) -> str:
  """
  Copies a file or directory to a temporary directory.

  Args:
    filename: File or directory path to copy.

  Returns:
    Destination path inside the temporary directory.

  Raises:
    FileNotFoundError: If the source does not exist.
    OSError: If copy errors occur.
  """
  temp_dir = get_temp_dir()
  if os.path.isdir(filename):
    name = os.path.basename(os.path.normpath(filename))
    dest = os.path.join(temp_dir, name)
    shutil.copytree(name, dest)
    return dest
  dest = os.path.join(temp_dir, os.path.basename(filename))
  shutil.copy2(filename, dest)
  return dest



# -*- coding: utf-8 -*-
# Catslap
# Author: Luis A. González
# MIT License (view LICENSE file)
# Copyright (c) 2026

# OOXML Python API by Luis Alberto González (SIA)
from catslap.base import utils as common
from catslap.utils import zip as zip_util
from catslap.utils import file as file_util


ZIPPED_EXTENSIONS = ['.docx', '.xlsx', '.pptx', '.zip']

class DocumentException(Exception):
  """
  Base exception for document processing.
  """
  pass


class Document:
  """
  Base class for OOXML documents.

  Attributes:
    filename: Input file path.
    value_resolver: Value resolution function.
    repeating_resolver: Repetition resolution function.
    is_zipped: Whether the document is a ZIP package.
    tempdir: Temporary working path.
    default_params: Default parameters.
    default_value_resolver: Default value resolver.
    default_repeat_resolver: Default repeat resolver.
    access_ok_param_list: List of successfully resolved parameters.
    access_err_param_list: List of unresolved parameters.
    test_mode: If True, avoids final writing.
  """
  def __init__(self, filename: str):
    """
    Creates the base document.

    Args:
      filename: Template file name to render.

    Raises:
      FileNotFoundError: If the file does not exist.
      OSError: If copy or extraction fails.
    """
    self.filename = filename
    self.value_resolver = None
    self.repeating_resolver = None
    extension = file_util.get_extension(filename).lower()
    is_dir = file_util.is_directory(filename)
    self.is_zipped = not is_dir and extension in ZIPPED_EXTENSIONS
    if self.is_zipped:
      self.tempdir = zip_util.extract_all(self.filename)
    else:
      self.tempdir = file_util.copy_to_temp_dir(filename)
    self.default_params = {}
    self.default_value_resolver = common.dict_value_resolver(self.default_params)
    self.default_repeat_resolver = common.dict_repeat_resolver(self.default_params)
    self.access_ok_param_list = []
    self.access_err_param_list = []
    self.test_mode = False
    self.config_params = {}

  def set_config_params(self, config: dict):
    self.config_params = config

  def get_document_bytes(self) -> bytes:
    """
    Returns the bytes of the generated document.

    Returns:
      Bytes of the generated document.

    Raises:
      OSError: If temporary read/write fails.
    """
    if self.is_zipped:
      zipfilename = file_util.get_temp_file(None, '.zip')
      try:
        zip_util.zip_directory(zipfilename, self.tempdir)
        return file_util.read_bytes(zipfilename)
      finally:
        file_util.remove_file(zipfilename)
    else:
      return file_util.read_bytes(self.tempdir)

  def close(self):
    """
    Closes the temporary resource (removes temp files/directories).

    Raises:
      OSError: If the temp resource cannot be removed.
    """
    if self.tempdir:
      if self.is_zipped:
        file_util.remove_dir_tree(self.tempdir)
      else:
        file_util.remove_file(self.tempdir)
      self.tempdir = None

  def create_doc_with_resolvers(self, output_file: str, __value_resolver, __repeating_resolver):
    """
    Generates a document using explicit resolvers.

    Args:
      output_file: Output path.
      __value_resolver: Value resolver.
      __repeating_resolver: Repeat resolver.

    Raises:
      OSError: If writing fails.
    """
    data = self.get_bytes_with_resolvers(__value_resolver, __repeating_resolver)
    file_util.write_bytes(output_file, data)

  def get_bytes_with_resolvers(self,  __value_resolver, __repeating_resolver) -> bytes:
    """
    Generates document bytes using explicit resolvers.

    Args:
      __value_resolver: Value resolver.
      __repeating_resolver: Repeat resolver.

    Returns:
      Bytes of the generated document.
    """
    self.value_resolver = __value_resolver
    self.repeating_resolver = __repeating_resolver
    self.process_template(self.tempdir)
    return self.get_document_bytes()

  def test_with_json(self,  json: dict) -> tuple[list, list]:
    """
    Processes the document in test mode using JSON.

    Args:
      json: Value map.

    Returns:
      Tuple with OK and error parameter lists.
    """
    return self.test_with_resolvers(common.dict_value_resolver(json), common.dict_repeat_resolver(json))

  def test_with_resolvers(self,  __value_resolver, __repeating_resolver) -> tuple[list, list]:
    """
    Processes the document in test mode using resolvers.

    Args:
      __value_resolver: Value resolver.
      __repeating_resolver: Repeat resolver.

    Returns:
      Tuple with OK and error parameter lists.
    """
    self.test_mode = True
    self.value_resolver = __value_resolver
    self.repeating_resolver = __repeating_resolver
    self.process_template(self.tempdir)
    return self.access_ok_param_list, self.access_err_param_list

  def create_doc_with_json(self, output_file: str, json: dict):
    """
    Generates a document using JSON.

    Args:
      output_file: Output path.
      json: Value map.

    Raises:
      OSError: If writing fails.
    """
    data = self.get_bytes_with_json(json)
    file_util.write_bytes(output_file, data)

  def get_bytes_with_json(self, json: dict) -> bytes:
    """
    Generates document bytes using JSON.

    Args:
      json: Value map.

    Returns:
      Bytes of the generated document.
    """
    return self.get_bytes_with_resolvers(common.dict_value_resolver(json), common.dict_repeat_resolver(json))

  def resolve_text(self, var_row: int | None, value: str) -> str:
    """
    Resolves {{ }} placeholders inside a text.

    Args:
      var_row: Current row index (if applicable).
      value: Text with placeholders.

    Returns:
      Resolved text.
    """
    rtext = ''
    idx0 = 0
    idx1 = value.find('{{')
    if idx1 < 0:
      return value
    while idx1 >= 0:
      idx2 = value.find('}}', idx1)
      if idx2 <= idx1:
        break
      rtext += value[idx0:idx1]
      param = value[idx1 + 2:idx2]
      resolved = self.resolve_value(var_row, param)
      if resolved is not None:
        rtext = rtext + str(resolved)
      idx0 = idx2 + 2
      idx1 = value.find('{{', idx0)
    return rtext

  def resolve_value(self, row: int | None, param: str) -> any:
    """
    Resolves a parameter using configured resolvers.

    Args:
      row: Current row index.
      param: Expression/key to resolve.

    Returns:
      Resolved value, or '' if not found.
    """
    value = self.default_value_resolver(row, param)
    if value is None:
      value = self.value_resolver(row, param)
    if value is None:
      if param not in self.access_err_param_list:
        self.access_err_param_list.append(param)
      return ''
    if param not in self.access_ok_param_list:
      self.access_ok_param_list.append(param)
    return value

  def resolve_repeating(self, param: str) -> int:
    """
    Resolves the number of repetitions for an expression.

    Args:
      param: Expression to evaluate.

    Returns:
      Repetition count (>=0) or -1 if invalid.
    """
    value = self.default_repeat_resolver(param)
    if value is None or value < 0:
      value = self.repeating_resolver(param)
    if value is None:
      value = 0
    if value < 0:
      if param not in self.access_err_param_list:
        self.access_err_param_list.append(param)
      return -1
    if param not in self.access_ok_param_list:
      self.access_ok_param_list.append(param)
    return value

  def process_template(self, tempdir: str):
    """
    Processes the template (abstract method).

    Args:
      tempdir: Temporary working directory.
    """
    pass



# -*- coding: utf-8 -*-
# Catslap
# Author: Luis A. González
# MIT License (view LICENSE file)
# Copyright (c) 2026

import base64


UTF_8 = "UTF-8"
ISO_8859_1 = "iso-8859-1"


def from_base64(string: str) -> bytes:
  """
  Decodes a Base64 string to bytes.

  Args:
    string: Base64 string (ISO-8859-1 encoded).

  Returns:
    Decoded bytes.

  Raises:
    binascii.Error: If the Base64 string is invalid.
    ValueError: If input conversion fails.
  """
  return base64.standard_b64decode(bytes(string, ISO_8859_1))



# -*- coding: utf-8 -*-
# Catslap
# Author: Luis A. González
# MIT License (view LICENSE file)
# Copyright (c) 2026

from abc import abstractmethod

from catslap.utils import encoding as enc_util
from catslap.utils import file as file_util
from catslap.utils import text as text_util
from catslap.utils import types
from catslap.utils import utils



class XmlException(Exception):
  """
  Base XML processing exception.
  """
  pass


class XmlParserException(XmlException):
  """
  XML parser specific exception.
  """
  pass


CONFIG_PARAM_STRICT = "STRICT"  # -- opciones estrictas de XML
CONFIG_PARAM_HTML = "HTML"  # -- ignora los tags cerrados sin que hayan abierto previamente (True) o error (False)
CONFIG_PARAM_AUTO_END_TAGS = "AUTO_END_TAGS"  # -- tags que deben o pueden ser autocerrados (string list)
CONFIG_PARAM_PRETTY_OUTPUT = "PRETTY_OUTPUT"
CONFIG_PARAM_INCLUDE_DECL = "INCLUDE_DECL"
HTML_AUTO_END_TAGS = ['img', 'br', 'meta', 'input']


xmlWriterDefaultConfig = {
    CONFIG_PARAM_AUTO_END_TAGS: HTML_AUTO_END_TAGS,
    CONFIG_PARAM_PRETTY_OUTPUT: True,
    CONFIG_PARAM_INCLUDE_DECL: False
}

xmlParserDefaultConfig = {
    CONFIG_PARAM_AUTO_END_TAGS: HTML_AUTO_END_TAGS,
    CONFIG_PARAM_HTML: True,
    CONFIG_PARAM_STRICT: False,
}


class XmlWriter:
  """
  XML writer with pretty print support.
  """
  def __init__(self, config: dict | None = None):
    self.out = []
    self.config = types.merge_dicts(config, xmlWriterDefaultConfig) if isinstance(config,
                                                                                      dict) else xmlWriterDefaultConfig
    if self.config[CONFIG_PARAM_INCLUDE_DECL]:
      self.out.append('<?xml version="1.0" encoding="UTF-8"?>')
      if self.config[CONFIG_PARAM_PRETTY_OUTPUT]:
        self.out.append('\n')

  def write(self, content: str):
    """
    Writes XML/HTML content to the buffer.

    Args:
      content: Text, XmlElement, or list of elements to serialize.
    """
    if isinstance(content, list):
      for item in content:
        if isinstance(item, XmlElement):
          item.write(0, self)
      return
    if isinstance(content, XmlElement):
      content.write(0, self)
      return
    if not isinstance(content, str):
      return
    self.out.append(content)

  def __str__(self) -> str:
    return ''.join(self.out)


class XmlElement:

  def __init__(self):
    self.parent = None

  """
  XML/HTML element.
  """
  @abstractmethod
  def write(self, indent: int, writer: XmlWriter):
    """
    Serializes the element.

    Args:
      indent: Indentation level.
      writer: Output writer.
    """
    pass

  @abstractmethod
  def clone(self, deep: bool = True) -> "XmlElement":
    """
    Clones the element.

    Args:
      deep: If True, recursively clones children.

    Returns:
      Element copy.
    """
    pass

  def to_xml(self) -> str:
    """
    Returns the serialized XML of the element.

    Returns:
      XML string.
    """
    writer = XmlWriter()
    self.write(0, writer)
    return str(writer)

  @abstractmethod
  def to_json(self) -> any:
    """
    Converts the element to a JSON representation.

    Returns:
      Equivalent JSON structure.
    """
    return None


class XmlText(XmlElement):
  """
  XML/HTML text node.
  """
  def __init__(self, content: str):
    self.content = content

  def append(self, text: str):
    """
    Appends text to the current node.

    Args:
      text: Text to concatenate.
    """
    self.content += text

  def write(self, indent: int, writer: XmlWriter):
    """
    Writes the text escaping entities.

    Args:
      indent: Indentation level (unused for text).
      writer: Output writer.
    """
    writer.write(XmlParser.escape_entities(self.content))

  def to_json(self) -> any:
    """
    Converts text to JSON.

    Returns:
      Dictionary with the 'text' key.
    """
    return {'text': self.content}

  def clone(self, deep: bool = True) -> XmlElement:
    """
    Clones the text node.

    Args:
      deep: Ignored (no children).

    Returns:
      New XmlText instance.
    """
    xml = XmlText(self.content)
    xml.parent = self.parent
    return xml


class XmlTag(XmlElement):
  """
  XML/HTML tag.
  """
  def __init__(self, tagname: str, attrs: dict | None = None):
    self.name = tagname
    self.attrs = attrs.copy() if isinstance(attrs, dict) else {}
    self.elements = []

  def clone(self, deep: bool = True) -> "XmlTag":
    """
    Clones the tag.

    Args:
      deep: If True, recursively clones children.

    Returns:
      Tag copy.
    """
    xml = XmlTag(self.name, self.attrs.copy())
    xml.parent = self.parent
    if deep:
      for item in self.elements:
        xml.add_element(item.clone())
    return xml

  def get_attr(self, attrname: str) -> str | None:
    """
    Gets an attribute by name.

    Args:
      attrname: Attribute name.

    Returns:
      Attribute value or None.
    """
    return self.attrs.get(attrname)

  def get_attr_int(self, attrname: str) -> int | None:
    """
    Gets an attribute and converts it to int if possible.

    Args:
      attrname: Attribute name.

    Returns:
      Integer or None.
    """
    value = self.get_attr(attrname)
    if value:
      value = value.strip()
      try:
        return int(value)
      except ValueError:
        pass
    return None

  def set_attr(self, attrname: str, attrvalue):
    """
    Sets an attribute.

    Args:
      attrname: Attribute name.
      attrvalue: Attribute value.
    """
    self.attrs[attrname] = attrvalue

  def remove_attr(self, attrname: str):
    """
    Removes an attribute if it exists.

    Args:
      attrname: Attribute name.
    """
    if self.attrs.get(attrname):
      del self.attrs[attrname]

  def add_attr(self, attrname: str, attrvalue: str | None):
    """
    Adds or updates an attribute.

    Args:
      attrname: Attribute name.
      attrvalue: Attribute value.
    """
    self.attrs[attrname] = attrvalue

  def add_attrs(self, attrs: dict):
    """
    Adds multiple attributes.

    Args:
      attrs: Attribute dictionary.
    """
    for attrname, attrvalue in attrs.items():
      self.attrs[attrname] = attrvalue

  def add_element(self, item: XmlElement) -> XmlElement:
    """
    Adds a child element.

    Args:
      item: XML element.

    Returns:
      Added element.

    Raises:
      XmlParserException: If the type is not XmlElement.
    """
    if not isinstance(item, XmlElement):
      raise XmlParserException("Invalid XML element type: " + str(type(item)))
    self.elements.append(item)
    item.parent = self
    return item

  def add_tag(self, tag: str or "XmlTag", attrs: dict | None = None) -> "XmlTag":
    """
    Creates or adds a child tag.

    Args:
      tag: Tag name or XmlTag.
      attrs: Optional attributes if tag is str.

    Returns:
      Added tag.

    Raises:
      XmlParserException: If the type is not XmlTag.
    """
    if isinstance(tag, str):
      tag = XmlTag(tag, attrs)
    if not isinstance(tag, XmlTag):
      raise XmlParserException("Invalid XML element type: " + str(type(tag)))
    self.add_element(tag)
    tag.parent = self
    return tag

  def add_tag_text(self, tag: str or "XmlTag", text: str|XmlText) -> "XmlTag":
    """
    Adds a child tag with text.

    Args:
      tag: Tag name or XmlTag.
      text: Text or XmlText.

    Returns:
      Created tag.
    """
    tag = self.add_tag(tag)
    tag.add_text(text)
    return tag

  def set_text(self, text: str|XmlText) -> XmlText | None:
    """
    Replaces content with a single text node.

    Args:
      text: Text or XmlText.

    Returns:
      Created text node.
    """
    self.elements = []
    return self.add_text(text)

  def add_text(self, text: str|XmlText) -> XmlText | None:
    """
    Adds text as a child of the tag.

    Args:
      text: Text or XmlText.

    Returns:
      Added text node.
    """
    if isinstance(text, XmlText):
      return self.add_element(text)
    if text is None:
      text = ''
    text = str(text)
    tags = self.elements
    text = XmlParser.resolve_entities(text)
    last_tag = tags[len(tags) - 1] if len(tags) > 0 else None
    if last_tag and isinstance(last_tag, XmlText):
      last_tag.append(text)
      return last_tag
    return self.add_element(XmlText(text))

  def write(self, indent: int, writer: "XmlWriter"):
    """
    Writes the tag and its content to the writer.

    Args:
      indent: Indentation level.
      writer: Output writer.
    """
    has_indent = writer.config[CONFIG_PARAM_PRETTY_OUTPUT]
    if has_indent:
      writer.write(utils.repeat(' ', indent * 2))
    writer.write('<' + self.name)
    if len(self.attrs) > 0:
      for attrname, attrvalue in self.attrs.items():
        writer.write(' ' + attrname)
        if attrvalue is not None:
          writer.write('="' + XmlParser.escape_attr_value(attrvalue) + '"')
    if len(self.elements) == 0:
      auto_end_tags = writer.config[CONFIG_PARAM_AUTO_END_TAGS]
      if auto_end_tags and self.name in auto_end_tags:
        writer.write('>')
      else:
        writer.write('/>')
      if has_indent:
        writer.write('\n')
      return
    writer.write('>')
    if len(self.elements) == 1 and isinstance(self.elements[0], XmlText):
      self.elements[0].write(indent + 1, writer)
      writer.write('</' + self.name + '>')
      if has_indent:
        writer.write('\n')
      return
    if has_indent:
      writer.write('\n')
    for tag in self.elements:
      tag.write(indent + 1, writer)
    if has_indent:
      writer.write(utils.repeat(' ', indent * 2))
    writer.write('</' + self.name + '>')
    if has_indent:
      writer.write('\n')

  def to_json(self) -> dict:
    """
    Converts the tag to JSON.

    Returns:
      Dictionary with name, attributes, and elements.
    """
    jmap = {'tag': self.name}
    if len(self.attrs) > 0:
      jmap['attrs'] = self.attrs
    if len(self.elements) > 0:
      tags = []
      for elem in self.elements:
        tags.append(elem.to_json())
      jmap['tags'] = tags
    return jmap

  def get_tag_path(self, paths: list, mandatory: bool = True) -> any:
    """
    Gets a tag by walking a name path.

    Args:
      paths: List of tag names (supports wildcards with '*').
      mandatory: If True, raises if missing.

    Returns:
      Found tag or None.

    Raises:
      XmlParserException: If a required tag is not found.
    """
    tag = self
    for item in paths:
      idx = item.find('*')
      if idx < 0:
        tag = tag.get_tag(item, mandatory)
        if tag is None:
          return None
        continue
      selected = None
      starts = item[0:idx]
      ends = item[idx + 1:]
      for elem in tag.elements:
        if not isinstance(elem, XmlTag):
          continue
        if (starts != '' and elem.name.startswith(starts)) or \
        (ends != '' and elem.name.endswith(ends)):
          selected = elem
          break
      if not selected:
        raise XmlParserException(f"Tag '{item}' nor found")
      tag = selected
    return tag

  def clear_tags(self):
    """
    Removes all children from the tag.
    """
    self.elements = []

  def get_tag(self, tag_name: str | None = None, mandatory: bool = True) -> "XmlTag":
    """
    Gets the first child tag with the given name.

    Args:
      tag_name: Tag name or None for the first.
      mandatory: If True, raises if missing.

    Returns:
      Found tag or None.

    Raises:
      XmlParserException: If missing and mandatory is True.
    """
    def __raise_exception():
      ex_name = '<' + tag_name + '>' if tag_name is not None else 'XML'
      raise XmlParserException(ex_name + ' tag expected below <' + self.name + '> tag')

    elements = self.elements
    for element in elements:
      if not isinstance(element, XmlTag):
        continue
      if tag_name is None or tag_name == element.name:
        return element
    if mandatory:
      __raise_exception()
    return None

  def remove_tag(self, tag_name: str):
    """
    Removes the first child tag with the given name.

    Args:
      tag_name: Tag name to remove.
    """
    idx = 0
    while idx < len(self.elements):
      elem = self.elements[idx]
      if not isinstance(elem, XmlTag):
        idx += 1
        continue
      if elem.name == tag_name:
        del self.elements[idx]
        return
      idx += 1

  def remove(self, elem0: XmlElement):
    """
    Removes a child element by reference.

    Args:
      elem0: Element to remove.
    """
    idx = 0
    while idx < len(self.elements):
      elem = self.elements[idx]
      if elem == elem0:
        del self.elements[idx]
        return
      idx += 1

  def remove_tags(self, tag_name: str):
    """
    Removes all child tags with the given name.

    Args:
      tag_name: Tag name to remove.
    """
    idx = 0
    while idx < len(self.elements):
      elem = self.elements[idx]
      if not isinstance(elem, XmlTag):
        idx += 1
        continue
      if elem.name == tag_name:
        del self.elements[idx]
        continue
      idx += 1

  def get_tag_text(self, tag_name: str, mandatory: bool = True) -> str | None:
    """
    Gets the text of the first child tag with the given name.

    Args:
      tag_name: Tag name.
      mandatory: If True, raises if missing.

    Returns:
      Tag text or None.
    """
    tag = self.get_tag(tag_name, mandatory)
    if tag is None:
      return None
    return tag.get_text()

  def get_tag_attr(self, tag_name: str, tag_attr: str, mandatory: bool = True) -> str | None:
    """
    Gets an attribute from the first child tag with the given name.

    Args:
      tag_name: Child tag name.
      tag_attr: Attribute name.
      mandatory: If True, raises if missing.

    Returns:
      Attribute value or None.
    """
    tag = self.get_tag(tag_name, mandatory)
    if tag is None:
      return None
    return tag.get_attr(tag_attr)

  def set_tag_text(self, tag_name: str, tag_value: str, mandatory: bool = True):
    """
    Sets text on the first child tag with the given name.

    Args:
      tag_name: Child tag name.
      tag_value: Text to set.
      mandatory: If True and missing, creates the tag.
    """
    tag = self.get_tag(tag_name, mandatory)
    if tag is None:
      tag = XmlTag(tag_name)
      self.add_tag(tag)
    tag.set_text(tag_value)

  def get_tags(self, tag_name: str | None = None) -> list:
    """
    Gets all child tags with the given name.

    Args:
      tag_name: Tag name or None for all.

    Returns:
      List of XmlTag.
    """
    tags = []
    elements = self.elements
    for block in elements:
      if not isinstance(block, XmlTag):
        continue
      if tag_name is not None and block.name != tag_name:
        continue
      tags.append(block)
    return tags

  def get_text(self) -> str:
    """
    Gets the text of the first text node.

    Returns:
      Node text.
    """
    text_node = self.get_text_node()
    return text_node.content

  def get_text_node(self, mandatory: bool = True) -> XmlText:
    """
    Gets the first text node.

    Args:
      mandatory: If True, raises if there is no text.

    Returns:
      XmlText (empty if missing and mandatory is False).
    """
    elements = self.elements
    if len(elements) == 0:
      return XmlText('')
    if not isinstance(elements[0], XmlText):
      if mandatory:
        raise XmlParserException('Text expected into <' + self.name + '> tag: ' + self.to_xml())
      return XmlText('')
    return elements[0]

  def get_inner_html(self) -> str:
    """
    Gets the inner HTML/XML of the tag.

    Returns:
      Serialized string of child elements.
    """
    writer = XmlWriter()
    for item in self.elements:
      item.write(0, writer)
    return str(writer)

class XmlParser:
  """
  XML/HTML parser with flexible configuration.
  """

  def __init__(self, config: dict = None):
    self.config = types.merge_dicts(config, xmlParserDefaultConfig) if isinstance(config, dict) else xmlParserDefaultConfig
    self.max_id = 0
    self.pathfile = None
    self.root_tag = None

  def parse_file(self, pathfile: str, roottag: str | None = None) -> XmlTag:
    """
    Parses an XML file and returns the root tag.

    Args:
      pathfile: XML file path.
      roottag: Expected root tag name.

    Returns:
      Parsed root tag.

    Raises:
      FileNotFoundError: If the file does not exist.
      XmlParserException: If XML is invalid or root mismatch.
    """
    self.pathfile = pathfile
    xml_content = text_util.trim(str(file_util.read_bytes(pathfile), enc_util.UTF_8))
    if xml_content.startswith('<?'):
      idx = xml_content.find('?>')
      if idx < 0:
        raise XmlParserException(": XML mark '...?>' not found")
      xml_content = text_util.trim(xml_content[idx + 2:])
    blocks = self.parse_text(xml_content)
    if len(blocks) < 1 or not isinstance(blocks[0], XmlTag):
      raise XmlParserException("None XML tag defined")
    block = blocks[0]
    if roottag:
      tag = block.name
      if tag != roottag:
        raise XmlParserException("XML root tag <" + roottag + " ...> was expected at start. Found <" + block.name + "> tag")
    self.root_tag = block
    return block

  def parse_text(self, text: str) -> list:
    """
    Parses XML/HTML from text.

    Args:
      text: XML/HTML text.

    Returns:
      List of parsed elements.

    Raises:
      XmlParserException: If the content is invalid.
    """
    self.root_tag = XmlTag('#root')
    self.__parse_tags(self.root_tag, text, 0, 0)
    return self.root_tag.elements

  def __parse_tags(self, parent_tag: XmlTag, row: str, idx: int, recurrent: int = 0) -> int:
    config_strict = self.config[CONFIG_PARAM_STRICT]
    config_html = self.config[CONFIG_PARAM_HTML]
    config_autoendtags = self.config[CONFIG_PARAM_AUTO_END_TAGS]
    opentag = parent_tag.name
    preserve = True
    while idx < len(row):
      idx0 = idx
      idx = row.find('<', idx0)
      if idx < 0:
      # -- texto como resto del doc
        text = row[idx0:]
        if not preserve:
          text = text_util.rtrim(text)
        _add_preserving_text(parent_tag, text, preserve)
        break
      if idx + 1 == len(row):
        if config_strict:
          XmlParser.__raise_xml_parse_exception(row, idx, "Invalid tag start mark at end of content")
          # -- texto como resto del doc (finaliza en <)
        text = row[idx0:]
        _add_preserving_text(parent_tag, text, preserve)
        break
        # -- texto hasta siguiente ...<tag> o ...</tag>
      text = row[idx0:idx]
      endtag = row[idx + 1] == '/'
      if endtag:
        idx += 1
        # -- caso especial '< ' o '</ '
      if not text_util.is_alpha(row[idx + 1]):
        idx += 1
        continue
        # -- toma el nombre del tag
      idx += 1
      idx0 = idx
      while idx < len(row) and row[idx] != '>' and row[idx] != ' ' and row[idx] != '/':
        idx += 1
      tag_name = row[idx0:idx]
      if tag_name == '':
        XmlParser.__raise_xml_parse_exception(row, idx0, "Empty tag name: <?>")
      idx2 = row.find(">", idx)
      if idx2 < 0:
        XmlParser.__raise_xml_parse_exception(row, idx0, f"End of tag mark expected: <{tag_name}...?")
      tag = XmlTag(tag_name)
      # -- parse atributos
      self.__parse_attrs(tag, row, idx, idx2) if not endtag else None
      autoendtag = row[idx2 - 1] == '/'
      if endtag and autoendtag:
        XmlParser.__raise_xml_parse_exception(row, idx0, f"Invalid autoend tag for tag end:  </{tag_name}/?")
      idx2 += 1
      idx0 = idx2
      # -- final de tag del tag de entrada, retorna
      if endtag and tag_name == opentag:
        _add_preserving_text(parent_tag, text, preserve)
        return idx0
      if not preserve:
        text = text_util.ltrim(text)
      if tag_name in config_autoendtags or autoendtag:
        _add_preserving_text(parent_tag, text, preserve)
        idx = idx0
        parent_tag.add_tag(tag)
        continue
      _add_preserving_text(parent_tag, text, preserve)
      parent_tag.add_tag(tag)
      if endtag:
        if config_strict and not config_html:
          XmlParser.__raise_xml_parse_exception(row, idx0, f"Invalid tag end </{tag_name}> without tag start. Expected </{opentag}>")
        idx = idx0
        # -- ignora tag
        continue
      idx = self.__parse_tags(tag, row, idx0, recurrent + 1)
    return idx

  def __parse_attrs(self, tag: XmlTag, row: str, idx: int, idx2: int):
    config_strict = self.config[CONFIG_PARAM_STRICT]
    tag_name = tag.name
    while idx < idx2:
    # -- ignora espacios
      while idx < idx2 and row[idx] in [' ', '\t']:
        idx += 1
        # -- coge nombre de attributo
      idx0 = idx
      while idx < idx2 and row[idx] not in [' ', '\t', '=', '/', '>']:
        idx += 1
      attr_name = row[idx0:idx]
      if attr_name == '':
        break
      if attr_name.find('"') >= 0 or attr_name.find("'") >= 0:
        XmlParser.__raise_xml_parse_exception(row, idx, f"Invalid attr name: <{tag_name} {attr_name}...")
      while idx < idx2 and row[idx] in [' ', '\t']:
        idx += 1
        # -- ignora el igual y coge el valor
      if idx >= idx2 or row[idx] != '=':
        if config_strict:
          XmlParser.__raise_xml_parse_exception(row, idx, f"Equals sign expected for attr: <{tag_name} {attr_name}=?...")
        tag.add_attr(attr_name, None)
        continue
      idx += 1
      while idx < idx2 and row[idx] in [' ', '\t']:
        idx += 1
      quote = row[idx]
      if quote != '"' and quote != "'":
        XmlParser.__raise_xml_parse_exception(row, idx, f"Quote expected around attribute from tag. Found '{quote}' char: <{tag_name} {attr_name}=?>")
      idx += 1
      idx0 = idx
      while idx < idx2 and row[idx] != quote:
        idx += 1
      quote2 = row[idx]
      attr_value = row[idx0:idx]
      if quote2 != quote:
        XmlParser.__raise_xml_parse_exception(row, idx, f"Quote expected around attribute from tag. Found '{quote2}' char: <{tag_name} {attr_name}={quote}{attr_value}{quote2}? ...>")
        continue
      idx += 1
      tag.add_attr(attr_name, XmlParser.resolve_entities(attr_value))
      if attr_name == 'id':
        try:
          idvalue = int(attr_value)
          if idvalue > self.max_id:
            self.max_id = idvalue
        except ValueError:
          pass

  def write_file(self):
    """
    Writes the root XML to the associated file.

    Raises:
      OSError: If the file cannot be written.
    """
    if self.pathfile and self.root_tag:
      writer = XmlWriter({CONFIG_PARAM_INCLUDE_DECL: True})
      self.root_tag.write(0, writer)
      content = str(writer)
      file_util.write_bytes(self.pathfile, content.encode(enc_util.UTF_8))

  @staticmethod
  def __calc_number_of_line(row: str, idx: int):
    line = 1
    if idx >= len(row):
      idx = len(row) - 1
    while idx >= 0:
      if row[idx] == '\n':
        line += 1
      idx -= 1
    return line

  @staticmethod
  def __raise_xml_parse_exception(row: str, idx: int, message: str):
    line = XmlParser.__calc_number_of_line(row, idx)
    raise XmlParserException(f"[line: {line}] {message}")

  @staticmethod
  def get_xml(blocks: list | None, config: dict | None = None) -> str:
    """
    Serializes elements to XML without pretty print.

    Args:
      blocks: List of elements.
      config: Additional configuration.

    Returns:
      Generated XML.
    """
    return XmlParser.__get_xml(blocks, types.merge_dicts({
      CONFIG_PARAM_PRETTY_OUTPUT: False
    }, config))

  @staticmethod
  def get_outer_xml(tag: XmlTag, config: dict | None = None) -> str:
    """
    Serializes a full tag to XML.

    Args:
      tag: Tag to serialize.
      config: Additional configuration.

    Returns:
      Generated XML.
    """
    return XmlParser.__get_xml(tag, types.merge_dicts({
      CONFIG_PARAM_PRETTY_OUTPUT: False
    }, config))

  @staticmethod
  def get_pretty_xml(blocks: list | None, config: dict | None = None) -> str:
    """
    Serializes elements to XML with pretty print.

    Args:
      blocks: List of elements.
      config: Additional configuration.

    Returns:
      Generated XML.
    """
    return XmlParser.__get_xml(blocks, types.merge_dicts({
      CONFIG_PARAM_PRETTY_OUTPUT: True
    }, config))

  @staticmethod
  def __get_xml(blocks: XmlElement | list | None, xml_writer_config: dict | None) -> str:
    if blocks is None:
      return ''
    out = XmlWriter(xml_writer_config)
    if isinstance(blocks, XmlElement):
      blocks.write(0, out)
      return str(out)
    if isinstance(blocks, list):
      for block in blocks:
        block.write(0, out)
      return str(out)
    raise XmlException("Invalid Xml argument: " + str(type(blocks)))

  @staticmethod
  def escape_entities(row) -> str:
    """
    Escapes basic XML entities.

    Args:
      row: Input text.

    Returns:
      Escaped text.
    """
    if row is None:
      return ''
    if isinstance(row, int) or isinstance(row, float):
      row = str(row)
    if isinstance(row, str):
      row = row.replace('&', '&amp;')
      row = row.replace('<', '&lt;')
      row = row.replace('>', '&gt;')
      return row
    return ''

  @staticmethod
  def escape_attr_value(row) -> str:
    """
    Escapes entities in attribute values.

    Args:
      row: Attribute value.

    Returns:
      Escaped value.
    """
    if isinstance(row, int) or isinstance(row, float):
      row = str(row)
    if isinstance(row, str):
      row = row.replace('&', '&amp;')
      row = row.replace('"', '&quot;')
      return row
    return ''

  @staticmethod
  def resolve_entities(row) -> str:
    """
    Resolves basic XML entities.

    Args:
      row: Input text.

    Returns:
      Text with resolved entities.
    """
    if isinstance(row, int) or isinstance(row, float):
      row = str(row)
    if isinstance(row, str):
      row = row.replace('&#xD;', '\r')
      row = row.replace('&#xA;', '\n')
      row = row.replace('&lt;', '<')
      row = row.replace('&gt;', '>')
      row = row.replace('&quot;', '"')
      row = row.replace('&apos;', "'")
      row = row.replace('&nbsp;', ' ')
      row = row.replace('&amp;', '&')  # -- tiene que ser al final para que no haya auto-resoluciones
      return row
    return ''

  @staticmethod
  def compose_autoclosed_tag(tagname: str, attrs: dict):
    """
    Composes a self-closing tag.

    Args:
      tagname: Tag name.
      attrs: Tag attributes.

    Returns:
      Self-closing XML string.
    """
    out = '<' + tagname
    if not attrs or len(attrs) == 0:
      out += '/>'
      return out
    for key, value in attrs.items():
      out += ' ' + key + '="' + value + '"'
    out += '/>\r\n'
    return out


def _add_preserving_text(parent_tag: XmlTag, text: str | None, preserve: bool):
  if text is None:
    return
  if not preserve:
    text = text.replace('\t', ' ').replace('\n', ' ').replace('\r', ' ')
    text = text_util.trim(text.replace('  ', ' '))
  if text == '':
    return
  parent_tag.add_text(text)



# -*- coding: utf-8 -*-
# Catslap
# Author: Luis A. González
# MIT License (view LICENSE file)
# Copyright (c) 2026

# OOXML Python API by Luis Alberto González (SIA)

from catslap.base.relationships import Relationships
from catslap.base.types import ContentTypes
from catslap.utils.xml import XmlParserException, XmlTag

TAG_BU_FONT = 'a:buFont'
TAG_BU_AUTO_NUM = 'a:buAutoNum'

ATTR_TYPEFACE_OL = '+mj-lt'
ATTR_TYPEFACE_DOTTED = '•'


def create_paragraph(props: dict, content: str or list, relationships: Relationships, types: ContentTypes):
  """
  Creates a PowerPoint paragraph with properties and content.

  Args:
    props: Paragraph/run properties.
    content: Text or list of XmlTag runs.
    relationships: Document relationships.
    types: Document ContentTypes.

  Returns:
    Paragraph XmlTag.

  Raises:
    XmlParserException: If content is invalid.
  """
  para_tag = XmlTag('a:p')
  p_props_tag = XmlTag('a:pPr')
  align = props.get('align')
  if align:
    p_props_tag.add_attr('algn', align)
  ul_list = props.get('list')
  if ul_list:
    p_props_tag.add_attr('marL', 171450)
    p_props_tag.add_attr('indent', -171450)
    if ul_list == 'number-dot':
      p_props_tag.add_tag(XmlTag(TAG_BU_FONT, {'typeface': ATTR_TYPEFACE_OL}))
      p_props_tag.add_tag(XmlTag(TAG_BU_AUTO_NUM, {'type': 'arabicParenR'}))
    elif ul_list == 'number':
      p_props_tag.add_tag(XmlTag(TAG_BU_FONT, {'typeface': ATTR_TYPEFACE_OL}))
      p_props_tag.add_tag(XmlTag(TAG_BU_AUTO_NUM, {'type': 'arabicPeriod'}))
    elif ul_list == 'alpha-dot':
      p_props_tag.add_tag(XmlTag(TAG_BU_FONT, {'typeface': ATTR_TYPEFACE_OL}))
      p_props_tag.add_tag(XmlTag(TAG_BU_AUTO_NUM, {'type': 'alphaLcParenR'}))
    elif ul_list == 'alpha':
      p_props_tag.add_tag(XmlTag(TAG_BU_FONT, {'typeface': ATTR_TYPEFACE_OL}))
      p_props_tag.add_tag(XmlTag(TAG_BU_AUTO_NUM, {'type': 'alphaLcPeriod'}))
    else:
      p_props_tag.add_tag(XmlTag('a:buChar', {'char': ATTR_TYPEFACE_DOTTED}))
  para_tag.add_tag(p_props_tag)
  if isinstance(content, str):
    runprops = props.copy()
    run_tag = create_run(content, runprops, relationships, types)
    para_tag.add_tag(run_tag)
    return para_tag
  if isinstance(content, list):
    for item in content:
      para_tag.add_tag(item)
    return para_tag
  raise XmlParserException("str o list expected: " + str(type(content)))


def create_run(text: str, runprops: dict or None, relationships: Relationships, types: ContentTypes) -> XmlTag:
  """
  Creates a text run for PowerPoint.

  Args:
    text: Run text.
    runprops: Formatting properties.
    relationships: Document relationships.
    types: Document ContentTypes.

  Returns:
    Run XmlTag.
  """
  rpr_tag = XmlTag('a:rPr')
  url = None
  if runprops:
    url = runprops.get('link')
    if url and relationships:
      relationship = relationships.add_relationship_hyperlink(url)
      rpr_tag.add_tag(XmlTag('a:hlinkClick', {'r:id': relationship.rid}))
    if runprops.get('bold'):
      rpr_tag.add_attr('b', '1')
    if runprops.get('italic'):
      rpr_tag.add_attr('i', '1')
    if runprops.get('underline'):
      rpr_tag.add_attr('u', 'sng')
    if runprops.get('strike'):
      rpr_tag.add_attr('strike', 'sngStrike')
    if runprops.get('code'):  # -- <a:latin typeface="Courier" pitchFamily="2" charset="0"/>
      rpr_tag.add_tag(XmlTag('a:latin', {'typeface': 'Courier', 'pitchFamily': 2, 'charset': 0}))
      # -- tamaño de fuente (por defecto siempre 800 -> 8pt)
    size = runprops.get('size')
    if not size or not isinstance(size, int):
      size = 800
    elif size < 700:
      size = 700
    if size:
      rpr_tag.add_attr('sz', size)

    color = runprops.get('color')
    if color:
      if color.startswith('#'):
        color = color[1:]
      solig_tag = rpr_tag.add_tag(XmlTag('a:solidFill'))
      solig_tag.add_tag(XmlTag('a:srgbClr', {'val': color}))
  run_tag = XmlTag('a:r')
  run_tag.add_tag(rpr_tag)
  if text is None:
    text = ''
  t_tag = run_tag.add_tag(XmlTag('a:t', {'xml:space': 'preserve'}))
  t_tag.add_text(text)
  return run_tag



# -*- coding: utf-8 -*-
# Catslap
# Author: Luis A. González
# MIT License (view LICENSE file)
# Copyright (c) 2026

# OOXML Python API by Luis Alberto González (SIA)

from catslap.utils.xml import XmlParser, XmlParserException

WORD_CONTENT_TYPES_XMLNS = "http://schemas.openxmlformats.org/package/2006/content-types"

def add_dict(dmap: dict, key: str, value):
  """
  Adds a key to a dict if the value is truthy.

  Args:
    dmap: Destination dictionary.
    key: Key to add.
    value: Value to evaluate.
  """
  if value:
    dmap[key] = value

class Default:
    """
    Default entry for Content Types.
    """
    def __init__(self, attrs: dict):
        """
        Represents a Default entry in [Content_Types].xml.

        Args:
          attrs: Attributes of the Default tag.
        """
        self.extension = attrs.get('Extension')
        self.content_type = attrs.get('ContentType')

class Override:
    """
    Override entry for Content Types.
    """
    def __init__(self, attrs: dict):
        """
        Represents an Override entry in [Content_Types].xml.

        Args:
          attrs: Attributes of the Override tag.
        """
        self.part_name = attrs.get('PartName')
        self.content_type = attrs.get('ContentType')

class ContentTypes(XmlParser):
  """
  Manages the [Content_Types].xml file.

  Attributes:
    defaults: List of Default entries.
    overrides: List of Override entries.
  """
  def __init__(self, pathfile):
    super().__init__()
    self.defaults = []
    self.overrides = []
    block = self.parse_file(pathfile, 'Types')
    blocks = block.get_tags()
    for block in blocks:
      tag = block.name
      attrs = block.attrs
      if tag == 'Default':
        entry = Default(attrs)
        self.defaults.append(entry)
      elif tag == 'Override':
        entry = Override(attrs)
        self.overrides.append(entry)
      else:
        raise XmlParserException("XML tag <" + tag + " ...> not valid for this file")

  def add_default(self, extension: str, content_type: str) -> Default:
    """
    Adds or updates a Default entry.

    Args:
      extension: File extension.
      content_type: Associated Content-Type.

    Returns:
      Created or updated Default entry.
    """
    for entry in self.defaults:
      if entry.extension == extension:
        entry.content_type = content_type
        return entry
    attrs = {}
    add_dict(attrs, 'Extension', extension)
    add_dict(attrs, 'ContentType', content_type)
    self.defaults.append(Default(attrs))
